#include "reltunnel.h"

u_int32_t curr_SEQ = SEQ_BEGIN ;


/* build_packet - takes header information and data and builds
 * 		  a reltunnel packet from it, returning the
 * 		  entire packet and an extra copy of the header
 * 		  seperately.
 * 
 * 	Returns packet length
 * 
 * (This function cant fail, yet, so no error checking is needed by the caller)
 */

int build_packet (const u_int8_t id, const u_int8_t rid, const u_int8_t flags,
		  const u_int32_t seq, const u_int8_t datalen,
		  const u_char *data, struct relhdr *rhdr, 
		  u_char *packet) 
{
   struct relhdr pkt ;
   
   
   
   pkt.id = id ;
   pkt.rid = rid ;
   pkt.flags = flags ;
   pkt.seq = seq ;
   pkt.hdr_cksum = 0 ;
   pkt.data_len = datalen ;
   pkt.data_cksum = 0 ;
   
   pkt.hdr_cksum = (u_int16_t) in_cksum((u_short *)&pkt, RELHDR_LEN) ;
   pkt.data_cksum = (u_int16_t) in_cksum((u_short *)data, datalen) ;


   memcpy(packet, &pkt, RELHDR_LEN) ; 
   packet+=RELHDR_LEN ; /* push the pointer to the data offset */
   memcpy(packet, data, datalen) ; /* copy application level data into the packet */
   packet-=RELHDR_LEN ; /* rewind it */
  
	           
   memcpy(rhdr, &pkt, RELHDR_LEN) ;
   
   return RELHDR_LEN + datalen ; /* return packet length, rhdr and packet */
}



/* explode_packet - like build_packet(), but reversed: 
 *    		    takes a built packet and parses it,
 * 		    returning the header, data, length, etc.
 * 
 * 		    Returns -1 on header cksum mismatch.
 * 		    Returns -2 on data cksum mismatch.
 * 		    Returns -3 for all other errors.
 * 		    Returns data length on success (>=0)
 */

int explode_packet (const u_char *packet, const int packet_len,
		    struct relhdr *pkthdr, u_char *data)
{
   struct relhdr testhdr ;

   u_int16_t test_hdrcksum = 0 , save_hdrcksum ;
   u_int16_t test_datacksum = 0, save_datacksum ;
   

   
   
   /* verify packet consistency */
   memcpy(&testhdr, packet, RELHDR_LEN) ;
   fprintf(stderr, "explode_packet:\t\ttesthdr.hdr_cksum=%u .seq=%u\n",
	   testhdr.hdr_cksum,
	   testhdr.seq
	   ) ;
   
   memcpy(&save_hdrcksum, &testhdr.hdr_cksum, sizeof(save_hdrcksum)) ;
   memcpy(&save_datacksum, &testhdr.data_cksum, sizeof(save_datacksum)) ;
   

   /* set hdr_cksum and data_cksum to 0 before re-calculating them */
   testhdr.hdr_cksum = 0 ;
   testhdr.data_cksum = 0 ; 

   test_hdrcksum = (u_int16_t) in_cksum((u_short *)&testhdr, RELHDR_LEN);

   /* header checksum mismatch. bail out */
   if (test_hdrcksum != save_hdrcksum) 
     return -1 ;
   
   
   packet+=RELHDR_LEN ; 
   test_datacksum = (u_int16_t) in_cksum((u_short *)packet, testhdr.data_len) ;
   /* data checksum mismatch, bail out. */
   if (test_datacksum != save_datacksum)
     return -2 ; 
   
   packet-=RELHDR_LEN ; 
   
   
   
   	packet+=RELHDR_LEN ; /* push the pointer to the data offset */
   memcpy(data, packet, testhdr.data_len) ; /* return data */
   	packet-=RELHDR_LEN ; /* and rewind it */

   memcpy(pkthdr, &testhdr, RELHDR_LEN) ; /* return pkthdr */

   return testhdr.data_len ;
}
   
 
/* __rel_send_packet  - encapsulates *data in a reltunnel packet
 * 			and sends it on socket s
 *			
 * 			Also verifies that our last packet
 * 			was succesfully received and ACKed
 */

int __rel__send_packet (int s, u_char *data, int data_len)
{
   int r = 0 ;
   int ok = 0 ;
   int b = 0 ; 
   u_char *packet ; 
   
   packet = (u_char *) malloc(data_len + RELHDR_LEN) ;
   
   /* verify that we're allowed to send this packet */
   ok = __rel__ok_to_send(curr_SEQ - SEQ_INCR) ;
   if (ok < 1) 
     return -1 ; 
   
     
   b = build_packet(conf.id, 
		    conf.rid, 
		    PSH,
		    curr_SEQ,
		    data_len,
		    data,
		    NULL,
		    packet
		    ) ;
		    
   
   if (b<1) {
      fprintf(stderr, 
	      "__rel__send_packet:\tbuild_packet() failed, "
	      "returned: %i\n",
	      b) ;
      return -1 ;
   }
   
   
   
    
   switch(sockets[s].type) {
    case IPPROTO_TCP:
      r = send(s, packet, b, 0) ;
      break ;
      
    default:
      fprintf(stderr, 
	      "__rel__send_packet:\tDont know how to send"
	      " on socket type %i for socket %i. Aborted\n",
	      sockets[s].type,
	      s);
      return -1 ;
      break ;
   }
   

   if (r<0) {
      fprintf(stderr,
	      "__rel__send_packet:\tsend(?) failed for %i bytes"
	      " on socket %i\n",
	      b,
	      s);
      return -1 ;
   }
   

   /* send succeeded, update sendlog database and increment SEQ */
   sendlog[curr_SEQ].send_time = get_long_time__usec() ;
   memcpy(sendlog[curr_SEQ].packet, packet, b) ;
   sendlog[curr_SEQ].packet_len = b ; 
   sendlog[curr_SEQ].resends = 0 ;
   sendlog[curr_SEQ].flags = PSH ; 
   
   curr_SEQ++; 

   return r ; 
}


int __rel__ok_to_send (long seq)
{
   if (sendlog[seq].acked == 1)
     return 1 ;
   

   fprintf(stderr, 
	   "packet %i not ACKed\n",
	   seq) ;
   
   return 0 ;
   
}


void __rel__acked_packet (long seq)
{

   sendlog[seq].send_time = 0 ;
   sendlog[seq].packet = NULL ;
   sendlog[seq].packet_len = 0 ;
   sendlog[seq].resends = 0 ;
   sendlog[seq].flags = 0 ;
   
}

